"""
Access to external tables of modular forms and elliptic curves
"""

#*****************************************************************************
#       MANIN: Modular Forms Software, (c) William Stein, 2004
#
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import os

import sage.misc.db as db
import sage.misc.errors as errors
import sage.misc.misc as misc
from sage.rings.arith import factor
import sage.pari.pari_pyx

PATH = "%s/src/data/cremona/"%os.environ["SAGE_HOME"]

class CremonaCurve:
    def __init__(self, conductor, iso_class, number,
                       a_invariants, rank, torsion, tamagawa):
        self.conductor = conductor
        self.iso_class = iso_class
        self.number = number
        self.a_invariants = a_invariants
        self.rank = rank
        self.torsion = torsion
        self.tamagawa = tamagawa   # {p: (p, cp), ...}

    def __repr__(self):
        s = "%s%s%s: invs=%s, rank=%s, tors=%s, cp=%s"%(\
            self.conductor, self.iso_class, self.number, self.a_invariants,
            self.rank, self.torsion, self.tamagawa)
        return s

    def tamagawa_product(self):
        x = 1
        for cp in self.tamagawa.itervalues():
            x *= cp
        return x

    def traces(self,n):
        cmd = ("ellan(ellinit(%s),%s)"%(self.a_invariants,n)).replace("L","")
        return eval(sage.pari.pari_pyx(cmd))
    

def create_db(file, start, stop):
    if not os.path.exists(PATH):
        os.makedirs(PATH)
    level = []
    clss = []
    N_last = 0
    first = True
    for line in open(file).readlines():
        X = line.split()
        if len(X) < 6:
            continue
        N, iso, num, ainvs, rank, tor = line.split()
        N = int(N)
        if N < start:
            continue
        if N > stop:
            return
        if N != N_last and N > 0:
            print "Saving level", N
            db.save(level, "%s/%s"%(PATH, N_last), gzip=True)
            level = []; clss = []
            level.append(clss)
            N_last = N
        if N_last == 0:
            N_last = N
        num = int(num)
        ainvs = eval(ainvs)
        rank = int(rank)
        tor = int(tor)
        P = [p for p, _ in factor(N)]
        cp = {}
        for p in P:
            cmd = "elllocalred(ellinit(%s),%s)"%(ainvs, p)
            cmd = cmd.replace("L","")
            x = eval(sage.pari.pari_pyx.eval(cmd))
            cp[p] = x[3]
        E = CremonaCurve(conductor = N,
                         iso_class = iso,
                         number = num,
                         a_invariants = ainvs,
                         rank = rank,
                         torsion = tor,
                         tamagawa = cp)

        if num == 1 and len(clss) > 0:
            clss = []
            level.append(clss)
        clss.append(E)
    db.save(level, "%s/%s"%(PATH, N), gzip=True)
        
def curves(conductor):
    """
    Data in the database about given conductor.
    """
    file = "%s/%s"%(PATH,conductor)
    if not os.path.exists(file+".gz"):
        return []
    return db.load(file, gzip=True)


def curve(label):
    i=0
    while i<len(label) and label[i]>='0' and label[i]<='9':
        i += 1
    j=i+1
    if j>len(label):
        label += "A"
    while j<len(label) and not (label[j]>='0' and label[j]<='9'):
        j += 1
    if j>=len(label):
        label += "1"
    level, iso, num = int(label[:i]), label[i:j], int(label[j:])
    iso = iso.upper()
    iso = (ord(iso) - ord('A')) + 26*(len(iso)-1)
    return curves(level)[iso][num-1]
